"""Keeper module to call keeper-contracts."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import logging
import os

from eth_utils import big_endian_to_int

from ocean_keeper.agreements.agreement_manager import AgreementStoreManager
from ocean_keeper.conditions.access import AccessSecretStoreCondition
from ocean_keeper.conditions.compute_execution import ComputeExecutionCondition
from ocean_keeper.conditions.condition_manager import ConditionStoreManager
from ocean_keeper.conditions.escrow_reward import EscrowRewardCondition
from ocean_keeper.conditions.hash_lock import HashLockCondition
from ocean_keeper.conditions.lock_reward import LockRewardCondition
from ocean_keeper.conditions.sign import SignCondition
from ocean_keeper.conditions.threshold import ThresholdCondition
from ocean_keeper.conditions.white_listing import WhitelistingCondition
from ocean_keeper.didregistry import DIDRegistry
from ocean_keeper.dispenser import Dispenser
from ocean_keeper.generic_contract import GenericContract
from ocean_keeper.templates.access_secret_store_template import EscrowAccessSecretStoreTemplate
from ocean_keeper.templates.compute_execution_template import EscrowComputeExecutionTemplate
from ocean_keeper.templates.template_manager import TemplateStoreManager
from ocean_keeper.token import Token
from ocean_keeper.utils import (add_ethereum_prefix_and_hash_msg, generate_multi_value_hash,
                                split_signature)
from ocean_keeper.wallet import Wallet
from ocean_keeper.web3.signature import SignatureFix
from ocean_keeper.web3_provider import Web3Provider


class Keeper(object):
    """The Keeper class aggregates all contracts in the Ocean Protocol node."""

    DEFAULT_NETWORK_NAME = 'development'
    _network_name_map = {
        1: 'Main',
        2: 'Morden',
        3: 'Ropsten',
        4: 'Rinkeby',
        42: 'Kovan',
        77: 'POA_Sokol',
        99: 'POA_Core',
        100: 'xDai',
        8995: 'nile',
        8996: 'spree',
        2199: 'duero',
        0xcea11: 'pacific'
    }

    def __init__(self, contract_names=None):
        self.network_name = Keeper.get_network_name(Keeper.get_network_id())
        self.accounts = Web3Provider.get_web3().eth.accounts
        self._contract_name_to_instance = {}
        if contract_names:
            for name in contract_names:
                try:
                    contract = GenericContract(name)
                    self._contract_name_to_instance = contract
                    setattr(self, name, contract)
                except (FileNotFoundError, KeyError, Exception) as err:
                    logging.debug(f'Contract artifact {name} is not found: {err}')

        self.dispenser = None
        if self.network_name != 'pacific':
            self.dispenser = Dispenser.get_instance()
        self.token = Token.get_instance()
        self.did_registry = DIDRegistry.get_instance()
        self.template_manager = TemplateStoreManager.get_instance()

        try:
            self.escrow_access_secretstore_template = EscrowAccessSecretStoreTemplate.get_instance()
        except (FileNotFoundError, KeyError, Exception) as err:
            self.escrow_access_secretstore_template = None
            logging.debug(f'Contract artifact {EscrowAccessSecretStoreTemplate.CONTRACT_NAME} is not found: {err}')

        try:
            self.escrow_compute_execution_template = EscrowComputeExecutionTemplate.get_instance()
        except (FileNotFoundError, KeyError, Exception) as err:
            self.escrow_compute_execution_template = None
            logging.debug(f'Contract artifact {EscrowComputeExecutionTemplate.CONTRACT_NAME} is not found: {err}')

        self.agreement_manager = AgreementStoreManager.get_instance()
        self.condition_manager = ConditionStoreManager.get_instance()
        self.sign_condition = SignCondition.get_instance()
        self.lock_reward_condition = LockRewardCondition.get_instance()
        self.escrow_reward_condition = EscrowRewardCondition.get_instance()
        self.access_secret_store_condition = AccessSecretStoreCondition.get_instance()
        self.compute_execution_condition = ComputeExecutionCondition.get_instance()
        self.hash_lock_condition = HashLockCondition.get_instance()
        self.threshold_condition = ThresholdCondition.get_instance()
        self.white_listing_condition = WhitelistingCondition.get_instance()
        contracts = [
            self.token,
            self.did_registry,
            self.template_manager,
            self.agreement_manager,
            self.condition_manager,
            self.sign_condition,
            self.lock_reward_condition,
            self.escrow_reward_condition,
            self.access_secret_store_condition,
            self.compute_execution_condition,
            self.hash_lock_condition,
            self.threshold_condition,
            self.white_listing_condition
        ]
        if self.dispenser:
            contracts.append(self.dispenser)
        self._contract_name_to_instance = {contract.name: contract
                                           for contract in contracts if contract}

    @staticmethod
    def get_instance(contract_names=None):
        """Return the Keeper instance (singleton)."""
        return Keeper(contract_names)

    @staticmethod
    def get_network_name(network_id):
        """
        Return the keeper network name based on the current ethereum network id.
        Return `development` for every network id that is not mapped.

        :param network_id: Network id, int
        :return: Network name, str
        """
        if os.environ.get('KEEPER_NETWORK_NAME'):
            logging.debug('keeper network name overridden by an environment variable: {}'.format(
                os.environ.get('KEEPER_NETWORK_NAME')))
            return os.environ.get('KEEPER_NETWORK_NAME')

        return Keeper._network_name_map.get(network_id, Keeper.DEFAULT_NETWORK_NAME)

    @staticmethod
    def get_network_id():
        """
        Return the ethereum network id calling the `web3.version.network` method.

        :return: Network id, int
        """
        return int(Web3Provider.get_web3().version.network)

    @staticmethod
    def sign_hash(msg_hash, account):
        """
        This method use `personal_sign`for signing a message. This will always prepend the
        `\x19Ethereum Signed Message:\n32` prefix before signing.

        :param msg_hash:
        :param account: Account
        :return: signature
        """
        wallet = Wallet(Web3Provider.get_web3(), account.key, account.password,
                        account.address)
        s = wallet.sign(msg_hash)
        return s.signature.hex()

    @staticmethod
    def ec_recover(message, signed_message):
        """
        This method does not prepend the message with the prefix `\x19Ethereum Signed Message:\n32`.
        The caller should add the prefix to the msg/hash before calling this if the signature was
        produced for an ethereum-prefixed message.

        :param message:
        :param signed_message:
        :return:
        """
        w3 = Web3Provider.get_web3()
        v, r, s = split_signature(w3, w3.toBytes(hexstr=signed_message))
        signature_object = SignatureFix(vrs=(v, big_endian_to_int(r), big_endian_to_int(s)))
        return w3.eth.account.recoverHash(message, signature=signature_object.to_hex_v_hacked())

    @staticmethod
    def personal_ec_recover(message, signed_message):
        prefixed_hash = add_ethereum_prefix_and_hash_msg(message)
        return Keeper.ec_recover(prefixed_hash, signed_message)

    @staticmethod
    def unlock_account(account):
        """
        Unlock the account.

        :param account: Account
        :return:
        """
        return Web3Provider.get_web3().personal.unlockAccount(account.address, account.password)

    @staticmethod
    def get_ether_balance(address):
        """
        Get balance of an ethereum address.

        :param address: address, bytes32
        :return: balance, int
        """
        return Web3Provider.get_web3().eth.getBalance(address, block_identifier='latest')

    @property
    def contract_name_to_instance(self):
        return self._contract_name_to_instance

    def get_contract(self, contract_name):
        contract = self.contract_name_to_instance.get(contract_name)
        if contract:
            return contract

        try:
            return GenericContract(contract_name)
        except (KeyError, Exception) as e:
            logging.error(f'Cannot load contract {contract_name}: {e}')
            return None

    def get_contract_by_address(self, contract_address):
        for contract in self._contract_name_to_instance.values():
            if contract.address == contract_address:
                return contract

        return None

    @staticmethod
    def generate_multi_value_hash(types, values):
        return generate_multi_value_hash(types, values)



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import os
import pathlib

from ocean_keeper import Keeper
from ocean_keeper.utils import get_account

PUBLISHER_INDEX = 1
CONSUMER_INDEX = 0


def get_resource_path(dir_name, file_name):
    base = os.path.realpath(__file__).split(os.path.sep)[1:-1]
    if dir_name:
        return pathlib.Path(os.path.join(os.path.sep, *base, dir_name, file_name))
    else:
        return pathlib.Path(os.path.join(os.path.sep, *base, file_name))


def init_ocn_tokens(ocn, account, amount=100):
    ocn.accounts.request_tokens(account, amount)
    Keeper.get_instance().token.token_approve(
        Keeper.get_instance().dispenser.address,
        amount,
        account
    )


def get_publisher_account():
    return get_account(0)


def get_consumer_account():
    return get_account(0)



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0
from ocean_keeper import Keeper
from ocean_keeper.templates.template_manager import TemplateStoreManager
from tests.resources.tiers import e2e_test


@e2e_test
def test_template():
    template_store_manager = TemplateStoreManager('TemplateStoreManager')
    assert template_store_manager.get_num_templates() == 2


@e2e_test
def test_default_templates():
    keeper = Keeper.get_instance()

    tm = keeper.template_manager
    access_template_name = tm.SERVICE_TO_TEMPLATE_NAME['access']
    compute_template_name = tm.SERVICE_TO_TEMPLATE_NAME['compute']
    access_id = tm.create_template_id(access_template_name)
    compute_id = tm.create_template_id(compute_template_name)
    # ids = [access_id, compute_id]
    assert tm.is_template_approved(access_id)
    assert tm.is_template_approved(compute_id)

    actor_ids = tm.get_template_actor_type_ids()
    for _id in actor_ids:
        type_value = tm.get_template_actor_type_value(_id)
        assert _id == tm.get_template_actor_type_id(type_value)
        assert 0 <= tm.get_template_actor_type_state(_id) <= 2

    template_values = tm.get_template(access_id)
    assert len(template_values.actor_type_ids) == 2
    assert tm.get_template_actor_type_value(template_values.actor_type_ids[0]) == 'provider'
    assert tm.get_template_actor_type_value(template_values.actor_type_ids[1]) == 'consumer'

    assert template_values.condition_types[0] == keeper.lock_reward_condition.address
    assert template_values.condition_types[1] == keeper.access_secret_store_condition.address
    assert template_values.condition_types[2] == keeper.escrow_reward_condition.address

    template_values = tm.get_template(compute_id)
    assert tm.get_template_actor_type_value(template_values.actor_type_ids[0]) == 'provider'
    assert tm.get_template_actor_type_value(template_values.actor_type_ids[1]) == 'consumer'

    assert template_values.condition_types[0] == keeper.lock_reward_condition.address
    assert template_values.condition_types[1] == keeper.compute_execution_condition.address
    assert template_values.condition_types[2] == keeper.escrow_reward_condition.address



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from ocean_keeper.conditions.condition_base import ConditionBase


class LockRewardCondition(ConditionBase):
    """Class representing the LockRewardCondition contract."""
    CONTRACT_NAME = 'LockRewardCondition'

    def fulfill(self, agreement_id, reward_address, amount, account):
        """
        Fulfill the lock reward condition.

        :param agreement_id: id of the agreement, hex str
        :param reward_address: ethereum account address, hex str
        :param amount: Amount of tokens, int
        :param account: Account instance
        :return:
        """
        return self._fulfill(
            agreement_id,
            reward_address,
            amount,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )

    def hash_values(self, reward_address, amount):
        """
        Hash of the values of the lock reward condition.

        :param reward_address: ethereum account address, hex str
        :param amount: Amount of tokens, int
        :return: hex str
        """
        return self._hash_values(reward_address, amount)



"""Test PaymentConditions contract."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from ocean_keeper.conditions.lock_reward import LockRewardCondition
from tests.resources.tiers import e2e_test


@e2e_test
def test_lock_reward_condition_contract():
    lock_reward_condition = LockRewardCondition('LockRewardCondition')
    assert lock_reward_condition
    assert isinstance(lock_reward_condition, LockRewardCondition), \
        f'{lock_reward_condition} is not instance of LockRewardCondition'



"""Keeper conditions."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from .access import AccessSecretStoreCondition
from .escrow_reward import EscrowRewardCondition
from .hash_lock import HashLockCondition
from .lock_reward import LockRewardCondition
from .sign import SignCondition



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from web3 import Web3

from ocean_keeper.web3.http_provider import CustomHTTPProvider


class Web3Provider(object):
    """Provides the Web3 instance."""
    _web3 = None

    @staticmethod
    def init_web3(keeper_url=None, provider=None):
        """
        One of `keeper_url` or `provider` is required. If `provider` is
        given, `keeper_url` will be ignored.

        :param keeper_url:
        :param provider:
        :return:
        """
        if not provider:
            assert keeper_url, 'keeper_url or a provider instance is required.'
            provider = CustomHTTPProvider(keeper_url)

        Web3Provider._web3 = Web3(provider)
        # Reset attributes to avoid lint issue about no attribute
        Web3Provider._web3.eth = getattr(Web3Provider._web3, 'eth')
        Web3Provider._web3.net = getattr(Web3Provider._web3, 'net')
        Web3Provider._web3.personal = getattr(Web3Provider._web3, 'personal')
        Web3Provider._web3.version = getattr(Web3Provider._web3, 'version')
        Web3Provider._web3.txpool = getattr(Web3Provider._web3, 'txpool')
        Web3Provider._web3.miner = getattr(Web3Provider._web3, 'miner')
        Web3Provider._web3.admin = getattr(Web3Provider._web3, 'admin')
        Web3Provider._web3.parity = getattr(Web3Provider._web3, 'parity')
        Web3Provider._web3.testing = getattr(Web3Provider._web3, 'testing')

    @staticmethod
    def get_web3(keeper_url=None, provider=None):
        """Return the web3 instance to interact with the ethereum client."""
        if Web3Provider._web3 is None:
            Web3Provider.init_web3(keeper_url, provider)
        return Web3Provider._web3

    @staticmethod
    def set_web3(web3):
        Web3Provider._web3 = web3



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0
import logging

from ocean_keeper.templates.template_base import TemplateBase

logger = logging.getLogger('escrowComputeExecutionTemplate')


class EscrowComputeExecutionTemplate(TemplateBase):
    """Class representing the EscrowComputeExecutionTemplate contract."""
    CONTRACT_NAME = 'EscrowComputeExecutionTemplate'

    def get_conditions_order(self):
        return ['', '', '']



import uuid

from ocean_keeper import Keeper
from ocean_keeper.web3_provider import Web3Provider
from tests.resources.helper_functions import get_publisher_account, get_consumer_account


def _new_id():
    return '0x' + uuid.uuid4().hex + uuid.uuid4().hex


def test_create_agreement():
    keeper = Keeper.get_instance()
    w3 = Web3Provider.get_web3()
    pub_acc = get_publisher_account()
    cons_acc = get_consumer_account()

    # register did
    did = _new_id()
    keeper.did_registry.register(did, w3.toBytes(0), f'/resolve/asset/{did}', pub_acc)

    # create agreement
    agr_id = _new_id()

    num_agreements = keeper.agreement_manager.get_num_agreements()

    template_mgr = keeper.template_manager
    template_id = template_mgr.create_template_id(template_mgr.SERVICE_TO_TEMPLATE_NAME['access'])
    template = keeper.template_manager.get_template(template_id)
    actor_types = [template_mgr.get_template_actor_type_value(_id) for _id in template.actor_type_ids]
    actor_type_map = {'provider': pub_acc.address, 'consumer': cons_acc.address}
    actors = [actor_type_map[actor_type] for actor_type in actor_types]

    access_id = keeper.access_secret_store_condition.generate_id(agr_id, ['bytes32', 'address'], [did, cons_acc.address]).hex()
    lock_id = keeper.lock_reward_condition.generate_id(agr_id, ['address', 'uint256'], [keeper.escrow_reward_condition.address, 100]).hex()
    reward_id = keeper.escrow_reward_condition.generate_id(
        agr_id,
        ['uint256', 'address', 'address', 'bytes32', 'bytes32'],
        [100, pub_acc.address, cons_acc.address, lock_id, access_id]
    ).hex()
    condition_types = template.condition_types
    condition_type_map = {
        keeper.access_secret_store_condition.address: (keeper.access_secret_store_condition, access_id),
        keeper.lock_reward_condition.address: (keeper.lock_reward_condition, lock_id),
        keeper.escrow_reward_condition.address: (keeper.escrow_reward_condition, reward_id)
    }
    ctm = condition_type_map
    condition_ids = [ctm[ct][1] for ct in condition_types]
    keeper.agreement_manager.create_agreement(
        agr_id,
        did,
        template_id,
        condition_ids,
        [0, 0, 0],
        [0, 0, 0],
        actors,
        cons_acc
    )

    # verify AgreementCreated event
    event = keeper.agreement_manager.subscribe_agreement_created(agr_id, 15, None, [], wait=True)
    assert event, f'agreement created event not found.'
    # verify agreement attributes on-chain
    agreement_values = keeper.agreement_manager.get_agreement(agr_id)
    assert agreement_values.template_id == template_id, f'templateId {template_id} does not match'
    assert agreement_values.condition_ids == condition_ids, f''
    assert agreement_values.owner == cons_acc.address

    # verify AgreementActorAdded events
    events = keeper.agreement_manager.get_event_filter_for_agreement_actor(
        cons_acc.address, from_block=agreement_values.block_number_updated).get_all_entries()
    assert events, f'no actor added event for consumer'
    assert events[0].args.actor == cons_acc.address
    # and for the publisher, use the old interface
    events = keeper.agreement_manager.get_event_filter_for_agreement_created(
        pub_acc.address, from_block=agreement_values.block_number_updated).get_all_entries()
    assert events, f'no actor added event for publisher'
    assert events[0].args.actor == pub_acc.address

    assert (num_agreements + 1) == keeper.agreement_manager.get_num_agreements()
    assert keeper.agreement_manager.get_agreement_did_owner(agr_id) == pub_acc.address
    assert keeper.agreement_manager.get_agreement_did_owner(agr_id[:-3] + '012') is None



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from contextlib import contextmanager
from inspect import signature

from ocean_keeper.web3_provider import Web3Provider


@contextmanager
def inject_dependencies(klass, *args, **kwargs):
    dependencies = kwargs.pop('dependencies', {})
    if 'dependencies' in signature(klass).parameters:
        kwargs['dependencies'] = dependencies

    to_restore = []

    def patch_provider(_object, _property, mock):
        to_restore.append((_object, _property, getattr(_object, _property)))
        setattr(_object, _property, mock)

    def maybe_patch_provider(_object, _property, name):
        if name in dependencies:
            patch_provider(_object, _property, dependencies[name])

    maybe_patch_provider(Web3Provider, '_web3', 'web3')
    try:
        yield klass(*args, **kwargs)
    finally:
        for (_object, _property, value) in to_restore:
            setattr(_object, _property, value)



"""Keeper module to call keeper-contracts."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import logging
from collections import namedtuple
from urllib.parse import urlparse

from web3 import Web3

from ocean_keeper.contract_base import ContractBase
from ocean_keeper.event_filter import EventFilter
from ocean_keeper.exceptions import OceanDIDNotFound

logger = logging.getLogger(__name__)

DIDRegisterValues = namedtuple(
    'DIDRegisterValues',
    ('owner', 'last_checksum', 'last_updated_by', 'block_number_updated', 'providers')
)


class DIDRegistry(ContractBase):
    """Class to register and update Ocean DID's."""
    DID_REGISTRY_EVENT_NAME = 'DIDAttributeRegistered'

    CONTRACT_NAME = 'DIDRegistry'

    def register(self, did, checksum, url, account, providers=None):
        """
        Register or update a DID on the block chain using the DIDRegistry smart contract.

        :param did: DID to register/update, can be a 32 byte or hexstring
        :param checksum: hex str hash of TODO
        :param url: URL of the resolved DID
        :param account: instance of Account to use to register/update the DID
        :param providers: list of addresses of providers to be allowed to serve the asset and play
            a part in creating and fulfilling service agreements
        :return: Receipt
        """
        if isinstance(did, bytes):
            did_source_id = did
        elif isinstance(did, str):
            try:
                did_source_id = Web3.toBytes(hexstr=did)
            except Exception as e:
                raise TypeError(
                    f'There is a problem with the did type, expecting a str with valid hex only, '
                    f'got {did}. Make sure to use only the id part of the '
                    f'did without the prefix: {e}')
        else:
            raise TypeError(f'Unrecognized `did` {did} of type {type(did)}. Expecting a 32 bytes '
                            f'in bytes type or hex str representation.')

        if not did_source_id:
            raise ValueError(f'{did} must be a valid DID to register')

        if not urlparse(url):
            raise ValueError(f'Invalid URL {url} to register for DID {did}')

        if checksum is None:
            checksum = Web3.toBytes(0)

        if not isinstance(checksum, bytes):
            raise ValueError(f'Invalid checksum value {checksum}, must be bytes or string')

        if account is None:
            raise ValueError('You must provide an account to use to register a DID')

        transaction = self._register_attribute(
            did_source_id, checksum, url, account, providers or []
        )
        receipt = self.get_tx_receipt(transaction)
        if receipt:
            return receipt.status == 1

        _filters = dict()
        _filters['_did'] = Web3.toBytes(hexstr=did)
        _filters['_owner'] = Web3.toBytes(hexstr=account.address)
        event = self.subscribe_to_event(self.DID_REGISTRY_EVENT_NAME, 15, _filters, wait=True)
        return event is not None

    def _register_attribute(self, did, checksum, value, account, providers):
        """Register an DID attribute as an event on the block chain.

        :param did: 32 byte string/hex of the DID
        :param checksum: checksum of the ddo, hex str
        :param value: url for resolve the did, str
        :param account: account owner of this DID registration record
        :param providers: list of providers addresses
        """
        assert isinstance(providers, list), ''
        return self.send_transaction(
            'registerAttribute',
            (did,
             checksum,
             providers,
             value),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )

    def get_block_number_updated(self, did):
        """Return the block number the last did was updated on the block chain."""
        return self.contract_concise.getBlockNumberUpdated(did)

    def get_did_owner(self, did):
        """
        Return the owner of the did.

        :param did: Asset did, did
        :return: ethereum address, hex str
        """
        return self.contract_concise.getDIDOwner(did)

    def add_provider(self, did, provider_address, account):
        """
        Add the provider of the did.

        :param did: the id of an asset on-chain, hex str
        :param provider_address: ethereum account address of the provider, hex str
        :param account: Account executing the action
        :return:
        """
        tx_hash = self.send_transaction(
            'addDIDProvider',
            (did,
             provider_address),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )
        return self.is_tx_successful(tx_hash)

    def remove_provider(self, did, provider_address, account):
        """
        Remove a provider

        :param did: the id of an asset on-chain, hex str
        :param provider_address: ethereum account address of the provider, hex str
        :param account: Account executing the action
        :return:
        """
        tx_hash = self.send_transaction(
            'removeDIDProvider',
            (did,
             provider_address),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )
        return self.is_tx_successful(tx_hash)

    def transfer_did_ownership(self, did, new_owner_address, account):
        """
        Transfer did ownership to an address.

        :param did: the id of an asset on-chain, hex str
        :param new_owner_address: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        tx_hash = self.send_transaction(
            'transferDIDOwnership',
            (did, new_owner_address),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )
        return self.is_tx_successful(tx_hash)

    def grant_permission(self, did, address_to_grant, account):
        """
        Grant access permission to an address.

        :param did: the id of an asset on-chain, hex str
        :param address_to_grant: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        tx_hash = self.send_transaction(
            'grantPermission',
            (did, address_to_grant),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )
        return self.is_tx_successful(tx_hash)

    def revoke_permission(self, did, address_to_revoke, account):
        """
        Revoke access permission to an address.

        :param did: the id of an asset on-chain, hex str
        :param address_to_revoke: ethereum account address, hex str
        :param account: Account executing the action
        :return: bool
        """
        tx_hash = self.send_transaction(
            'revokePermission',
            (did, address_to_revoke),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )
        return self.is_tx_successful(tx_hash)

    def get_permission(self, did, address):
        """
        Gets access permission of a grantee

        :param did: the id of an asset on-chain, hex str
        :param address: ethereum account address, hex str
        :return: true if the address has access permission to a DID
        """
        return self.contract_concise.getPermission(did, address)

    def is_did_provider(self, did, address):
        """
        Return true if the address is the Provider of the did.

        :param did: the id of an asset on-chain, hex str
        :param address: ethereum account address, hex str
        :return: bool
        """
        return self.contract_concise.isDIDProvider(did, address)

    def get_did_providers(self, did):
        """
        Return the list providers registered on-chain for the given did.

        :param did: hex str the id of an asset on-chain
        :return:
            list of addresses
            None if asset has no registerd providers
        """
        register_values = self.contract_concise.getDIDRegister(did)
        if register_values and len(register_values) == 5 and register_values[0]:
            # sanitize providers list, because if providers were removed they will
            # be replaced with null/None
            valid_providers = [a for a in register_values[4] if a is not None]
            register_values[4] = valid_providers
            return DIDRegisterValues(*register_values).providers

        return None

    def get_owner_asset_ids(self, address):
        """
        Get the list of assets owned by an address owner.

        :param address: ethereum account address, hex str
        :return:
        """
        block_filter = self._get_event_filter(owner=address)
        log_items = block_filter.get_all_entries(max_tries=5)
        did_list = []
        for log_i in log_items:
            did_list.append(log_i.args['_did'])

        return did_list

    def get_registered_attribute(self, did_bytes):
        """

        Example of event logs from event_filter.get_all_entries():
        [AttributeDict(
            {'args': AttributeDict(
                {'did': b'\x02n\xfc\xfb\xfdNM\xe9\xb8\xe0\xba\xc2\xb2\xc7\xbeg\xc9/\x95\xc3\x16\
                           x98G^\xb9\xe1\xf0T\xce\x83\xcf\xab',
                 'owner': '0xAd12CFbff2Cb3E558303334e7e6f0d25D5791fc2',
                 'value': 'http://localhost:5000',
                 'checksum': '0x...',
                 'updatedAt': 1947}),
             'event': 'DIDAttributeRegistered',
             'logIndex': 0,
             'transactionIndex': 1,
             'transactionHash': HexBytes(
             '0xea9ca5748d54766fb43fe9660dd04b2e3bb29a0fbe18414457cca3dd488d359d'),
             'address': '0x86DF95937ec3761588e6DEbAB6E3508e271cC4dc',
             'blockHash': HexBytes(
             '0xbbbe1046b737f33b2076cb0bb5ba85a840c836cf1ffe88891afd71193d677ba2'),
             'blockNumber': 1947})]

        """
        result = None
        did = Web3.toHex(did_bytes)
        block_number = self.get_block_number_updated(did_bytes)
        logger.debug(f'got blockNumber {block_number} for did {did}')
        if block_number == 0:
            raise OceanDIDNotFound(
                f'DID "{did}" is not found on-chain in the current did registry. '
                f'Please ensure assets are registered in the correct keeper contracts. '
                f'The keeper-contracts DIDRegistry address is {self.address}.'
                f'Also ensure that sufficient time has passed after registering the asset '
                f'such that the transaction is confirmed by the network validators.')

        block_filter = self._get_event_filter(did=did, from_block=block_number,
                                              to_block=block_number)
        log_items = block_filter.get_all_entries(max_tries=5)
        if not log_items:
            block_filter = self._get_event_filter(did=did, from_block=block_number - 1,
                                                  to_block=block_number + 1)
            log_items = block_filter.get_all_entries(max_tries=3)

        if log_items:
            log_item = log_items[-1].args
            value = log_item['_value']
            block_number = log_item['_blockNumberUpdated']
            result = {
                'checksum': log_item['_checksum'],
                'value': value,
                'block_number': block_number,
                'did_bytes': log_item['_did'],
                'owner': Web3.toChecksumAddress(log_item['_owner']),
            }
        else:
            logger.warning(f'Could not find {DIDRegistry.DID_REGISTRY_EVENT_NAME} event logs for '
                           f'did {did} at blockNumber {block_number}')
        return result

    def _get_event_filter(self, did=None, owner=None, from_block=0, to_block='latest'):
        _filters = {}
        if did is not None:
            _filters['_did'] = Web3.toBytes(hexstr=did)
        if owner is not None:
            _filters['_owner'] = Web3.toBytes(hexstr=owner)

        block_filter = EventFilter(
            DIDRegistry.DID_REGISTRY_EVENT_NAME,
            getattr(self.events, DIDRegistry.DID_REGISTRY_EVENT_NAME),
            from_block=from_block,
            to_block=to_block,
            argument_filters=_filters
        )
        return block_filter



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0
import os

import pytest
from ocean_keeper.contract_handler import ContractHandler
from ocean_keeper.wallet import Wallet
from ocean_keeper.web3_provider import Web3Provider
from web3 import HTTPProvider, Web3
from ocean_keeper.keeper import Keeper

from tests.resources.helper_functions import (
    get_consumer_account,
    get_publisher_account
)


def get_keeper_url():
    if os.getenv('KEEPER_URL'):
        return os.getenv('KEEPER_URL')
    return 'http://localhost:8545'


@pytest.fixture(autouse=True)
def setup_all():
    Web3Provider.init_web3(get_keeper_url())
    default_path = '~/.ocean/keeper-contracts/artifacts'
    keeper_path = os.environ.get('ARTIFACTS_PATH', default_path)
    ContractHandler.set_artifacts_path(os.path.expanduser(os.path.expandvars(keeper_path)))
    Keeper.get_instance()
    Wallet.reset_tx_count()


@pytest.fixture
def publisher_account():
    return get_publisher_account()


@pytest.fixture
def consumer_account():
    return get_consumer_account()


@pytest.fixture
def web3_instance():
    return Web3(HTTPProvider(get_keeper_url()))



"""Test DIDRegistry contract."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import secrets
import uuid

import pytest
from eth_utils import add_0x_prefix
from web3 import Web3

from ocean_keeper.didregistry import DIDRegistry
from tests.resources.helper_functions import get_consumer_account, get_publisher_account
from tests.resources.tiers import e2e_test


def new_did():
    return uuid.uuid4().hex + uuid.uuid4().hex


@e2e_test
def test_did_registry_contract():
    did_registry = DIDRegistry.get_instance()
    assert did_registry
    assert isinstance(did_registry, DIDRegistry)


@e2e_test
def test_did_registry_get_block_number_updated():
    did_registry = DIDRegistry.get_instance()
    test_id = secrets.token_hex(32)
    assert did_registry.get_block_number_updated(test_id) == 0


@e2e_test
def test_register():
    did_registry = DIDRegistry.get_instance()
    w3 = Web3
    _id = new_did()
    did_test = 'did:op:' + _id
    register_account = get_publisher_account()
    checksum_test = w3.sha3(text='checksum')
    value_test = 'http://localhost:5000'
    # register DID-> URL
    assert did_registry.register(
        _id, checksum_test, url=value_test, account=register_account
    ) is True

    with pytest.raises(Exception):
        did_registry.get_did_owner(did_test)

    assert did_registry.get_did_owner(
        add_0x_prefix(_id)
    ) == register_account.address


@e2e_test
def test_register_with_invalid_params():
    did_registry = DIDRegistry.get_instance()
    w3 = Web3
    did_test = new_did()
    checksum_test = w3.sha3(text='checksum')
    value_test = 'http://localhost:5000'
    # No checksum provided
    with pytest.raises(ValueError):
        did_registry.register(did_test, '', url=value_test, account=None)
    # Invalid checksum  provided
    with pytest.raises(ValueError):
        did_registry.register(did_test, did_test, url=value_test, account=None)

    # No account provided
    with pytest.raises(ValueError):
        did_registry.register(did_test, checksum_test, url=value_test, account=None)


@e2e_test
def test_providers():
    did_registry = DIDRegistry.get_instance()
    w3 = Web3
    did_test = new_did()
    asset_id = add_0x_prefix(did_test)
    register_account = get_publisher_account()
    consumer_account = get_consumer_account()
    checksum_test = w3.sha3(text='checksum')
    value_test = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')
    # register DID-> URL
    with pytest.raises(AssertionError):
        did_registry.register(
            did_test, checksum_test, url=value_test, account=register_account,
            providers=consumer_account.address
        )
    did_registry.register(
        did_test, checksum_test, url=value_test, account=register_account, providers=[test_address]
    )
    unknown_asset_id = add_0x_prefix(new_did())
    providers = did_registry.get_did_providers(unknown_asset_id)
    assert providers is None

    assert did_registry.is_did_provider(asset_id, register_account.address) is False

    providers = did_registry.get_did_providers(asset_id)
    assert len(providers) == 1 and providers[0] == test_address
    assert did_registry.is_did_provider(asset_id, test_address) is True

    removed = did_registry.remove_provider(asset_id, test_address, register_account)
    assert removed
    providers = did_registry.get_did_providers(asset_id)
    assert len(providers) == 0
    assert did_registry.is_did_provider(asset_id, test_address) is False

    did_registry.add_provider(asset_id, test_address, register_account)
    providers = did_registry.get_did_providers(asset_id)
    assert len(providers) == 1 and providers[0] == test_address
    assert did_registry.is_did_provider(asset_id, test_address) is True


def test_transfer_ownership():
    did_registry = DIDRegistry.get_instance()
    w3 = Web3
    did_test = new_did()
    asset_id = add_0x_prefix(did_test)
    register_account = get_publisher_account()
    consumer_account = get_consumer_account()
    checksum_test = w3.sha3(text='checksum')
    value_test = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    did_registry.register(did_test, checksum_test, url=value_test, account=register_account,
                          providers=[test_address])

    first_owner = did_registry.get_did_owner(did_test)
    assert first_owner == register_account.address

    did_registry.transfer_did_ownership(did_test, test_address, register_account)
    assert did_registry.get_did_owner(did_test) == test_address


def test_grant_permissions():
    did_registry = DIDRegistry.get_instance()
    w3 = Web3
    did_test = new_did()
    register_account = get_publisher_account()
    checksum_test = w3.sha3(text='checksum')
    value_test = 'http://localhost:5000'
    test_address = w3.toChecksumAddress('068ed00cf0441e4829d9784fcbe7b9e26d4bd8d0')

    did_registry.register(did_test, checksum_test, url=value_test, account=register_account,
                          providers=[test_address])

    assert not did_registry.get_permission(did_test, test_address)

    did_registry.grant_permission(did_test, test_address, register_account)

    assert did_registry.get_permission(did_test, test_address)

    did_registry.revoke_permission(did_test, test_address, register_account)

    assert not did_registry.get_permission(did_test, test_address)



"""Test the Account object."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from ocean_keeper.account import Account
from tests.resources.helper_functions import get_resource_path


def test_create_account():
    account = Account('0x213123123', 'pass', encrypted_key='0190209129092101920')
    assert isinstance(account, Account)
    assert account.address == '0x213123123'
    assert account.password == 'pass'
    assert account.key == '0190209129092101920'

    account = Account('0x213123123', 'pass', private_key='0190209129092101920')
    assert isinstance(account, Account)
    assert account.address == '0x213123123'
    assert account.password is None
    assert account.key == '0190209129092101920'

    address = '0x00bd138abd70e2f00903268f3db08f2d25677c9e'
    key_file = get_resource_path('data', 'publisher_key_file.json')
    with open(key_file) as f:
        encrypted_key = f.read()
    account = Account(address, 'pass', key_file=key_file)
    assert account.key_file == str(key_file)
    assert account.key == encrypted_key

    account = Account(address, 'pass', key_file=key_file,
                      encrypted_key='encrypted_key')
    assert account.key == 'encrypted_key'

    account = Account(address, 'pass', key_file=key_file,
                      encrypted_key=encrypted_key, private_key='privatekey')
    assert account.key == 'privatekey'



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import logging

from ocean_keeper.templates.template_base import TemplateBase

logger = logging.getLogger('escrowAccessSecretStoreTemplate')


class EscrowAccessSecretStoreTemplate(TemplateBase):
    """Class representing the EscrowAccessSecretStoreTemplate contract."""
    CONTRACT_NAME = 'EscrowAccessSecretStoreTemplate'

    def get_conditions_order(self):
        return ['', '', '']



"""Test AccessConditions contract."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0
import uuid

from ocean_keeper.conditions.access import AccessSecretStoreCondition
from tests.resources.helper_functions import get_consumer_account
from tests.resources.tiers import e2e_test


@e2e_test
def test_access_secret_store_condition_contract():
    access_secret_store_condition = AccessSecretStoreCondition('AccessSecretStoreCondition')
    assert access_secret_store_condition
    assert isinstance(access_secret_store_condition, AccessSecretStoreCondition), \
        f'{access_secret_store_condition} is not instance of AccessSecretStoreCondition'


@e2e_test
def test_check_permissions_not_registered_did():
    access_secret_store_condition = AccessSecretStoreCondition('AccessSecretStoreCondition')
    consumer_account = get_consumer_account()
    did_id = uuid.uuid4().hex + uuid.uuid4().hex
    assert not access_secret_store_condition.check_permissions(did_id, consumer_account.address)

# TODO Create test for check permission after access granted.



"""Keeper templates module."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from .access_secret_store_template import EscrowAccessSecretStoreTemplate
from .template_manager import TemplateStoreManager



class OceanKeeperContractsNotFound(Exception):
    """Raised when is not possible to find the keeper contracts abi."""


class OceanDIDNotFound(Exception):
    """Raised when a requested DID or a DID in the chain cannot be found."""


class OceanInvalidTransaction(Exception):
    """Raised when an on-chain transaction fail."""



__author__ = """OceanProtocol"""
__version__ = '0.3.3'

#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from .contract_base import ContractBase
from .didregistry import DIDRegistry
from .dispenser import Dispenser
from .keeper import Keeper
from .token import Token



"""Keeper agreements sub-module."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0
import logging
from collections import namedtuple

from eth_utils import add_0x_prefix
from web3.contract import ContractEvent
from web3.exceptions import MismatchedABI

from ocean_keeper import ContractBase
from ocean_keeper.event_filter import EventFilter
from ocean_keeper.web3_provider import Web3Provider

logger = logging.getLogger(__name__)


AgreementValues = namedtuple(
    'AgreementValues',
    ('did', 'owner', 'template_id', 'condition_ids', 'updated_by', 'block_number_updated')
)


class AgreementStoreManager(ContractBase):
    """Class representing the AgreementStoreManager contract."""
    CONTRACT_NAME = 'AgreementStoreManager'
    AGREEMENT_CREATED_EVENT = 'AgreementCreated'
    AGREEMENT_ACTOR_ADDED_EVENT = 'AgreementActorAdded'

    agreement_actor_added_event_abi = {
        "abi": [{
          "anonymous": False,
          "inputs": [
            {
              "indexed": True,
              "name": "agreementId",
              "type": "bytes32"
            },
            {
              "indexed": True,
              "name": "actor",
              "type": "address"
            }
          ],
          "name": "AgreementActorAdded",
          "type": "event"
    }]}

    def create_agreement(
            self,
            agreement_id,
            did,
            template_id,
            condition_ids,
            time_locks,
            time_outs,
            actors,
            account
    ):
        """
        Create a new agreement.
        The agreement will create conditions of conditionType with conditionId.
        Only "approved" templates can access this function.
        :param agreement_id:id of the agreement, hex str
        :param did: DID of the asset. The DID must be registered beforehand, bytes32
        :param template_id: hex str of bytes32 that is the id of an approved agreement template
        :param condition_ids: is a list of bytes32 content-addressed Condition IDs, bytes32
        :param time_locks: is a list of uint time lock values associated to each Condition, int
        :param time_outs: is a list of uint time out values associated to each Condition, int
        :param actors: list of ethereum addresses of each of the actors in this agreement. The
            order of actors should match the order defined in the template
        :return: bool
        """
        tx_hash = self.send_transaction(
            'createAgreement',
            (agreement_id,
             did,
             template_id,
             condition_ids,
             time_locks,
             time_outs,
             actors),
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key},
        )
        return self.is_tx_successful(tx_hash)

    def get_agreement(self, agreement_id):
        """
        Retrieve the agreement for a agreement_id.

        :param agreement_id: id of the agreement, hex str
        :return: the agreement attributes.
        """
        agreement = self.contract_concise.getAgreement(agreement_id)
        if agreement and len(agreement) == 6:
            agreement = AgreementValues(*agreement)
            did = add_0x_prefix(agreement.did.hex())
            cond_ids = [add_0x_prefix(_id.hex()) for _id in agreement.condition_ids]

            return AgreementValues(
                did,
                agreement.owner,
                add_0x_prefix(agreement.template_id.hex()),
                cond_ids,
                agreement.updated_by,
                agreement.block_number_updated
            )

        return None

    def get_agreement_did_owner(self, agreement_id):
        """Get the DID owner for this agreement with _id.

        :param agreement_id: id of the agreement, hex str
        :return: the DID owner associated with agreement.did from the DID registry.
        """
        return self.contract_concise.getAgreementDIDOwner(agreement_id)

    def get_num_agreements(self):
        """Return the size of the Agreements list.

        :return: the length of the agreement list, int
        """
        return self.contract_concise.getAgreementListSize()

    def subscribe_agreement_created(self, agreement_id, timeout, callback, args, wait=False,
                                    from_block='latest', to_block='latest'):
        """
        Subscribe to an AgreementCreated event matching the given `agreement_id`.

        :param agreement_id: hex str -- id of the agreement
        :param timeout: int -- time in seconds to wait for the event
        :param callback: function callback when the
        :param args:
        :param wait: bool -- if true block the listener until the event is fetched,
            else return immediately and process the event in the background
        :param from_block: int or None -- the block number to start searching for matching events
        :param to_block: int or None -- the block number to end searching for matching events
        :return: event (dict) if blocking is True and an event is received, otherwise returns None

        """
        logger.debug(
            f'Subscribing {self.AGREEMENT_CREATED_EVENT} event '
            f'with agreement id {agreement_id}.'
        )
        return self.subscribe_to_event(
            self.AGREEMENT_CREATED_EVENT,
            timeout,
            {'agreementId': Web3Provider.get_web3().toBytes(hexstr=agreement_id)},
            callback=callback,
            args=args,
            wait=wait,
            from_block=from_block,
            to_block=to_block
        )

    def _get_contract_agreement_actor_added_event(self):
        try:
            event = getattr(self.events, self.AGREEMENT_ACTOR_ADDED_EVENT, None)
        except MismatchedABI:
            event = None

        if not event:
            event = ContractEvent.factory(
                self.AGREEMENT_ACTOR_ADDED_EVENT,
                web3=Web3Provider.get_web3(),
                contract_abi=self.agreement_actor_added_event_abi['abi'],
                address=self.address,
                event_name=self.AGREEMENT_ACTOR_ADDED_EVENT)

        return event

    def get_event_filter_for_agreement_created(self, provider_address=None, from_block='latest',
                                               to_block='latest'):
        return self.get_event_filter_for_agreement_actor(provider_address, from_block, to_block)

    def get_event_filter_for_agreement_actor(self, actor_address, from_block='latest', to_block='latest'):
        """
        :param actor_address: hex str ethereum address of actor
        :param from_block: int or None
        :param to_block: int or None
        :return:
        """

        _filter = {}
        if actor_address:
            assert isinstance(actor_address, str)
            _filter['actor'] = actor_address

        event_filter = EventFilter(
            self.AGREEMENT_ACTOR_ADDED_EVENT,
            self._get_contract_agreement_actor_added_event(),
            _filter,
            from_block=from_block,
            to_block=to_block
        )
        event_filter.set_poll_interval(0.5)
        return event_filter



"""Accounts module."""

#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import logging
import os

logger = logging.getLogger('account')


class Account:
    """Class representing an account."""

    def __init__(self, address, password=None, key_file=None, encrypted_key=None, private_key=None):
        """
        Hold account address, password and either keyfile path, encrypted key or private key

        :param address: The address of this account
        :param password: account's password. This is necessary for decrypting the private key
            to be able to sign transactions locally
        :param key_file: str path to the encrypted private key file
        :param encrypted_key:
        :param private_key:
        """
        assert key_file or encrypted_key or private_key, \
            'Account requires one of `key_file`, `encrypted_key`, or `private_key`.'
        if key_file or encrypted_key:
            assert password, '`password` is required when using `key_file` or `encrypted_key`.'

        if private_key:
            password = None

        self.address = address
        self.password = password
        self._key_file = key_file
        if self._key_file and not encrypted_key:
            with open(self.key_file) as _file:
                encrypted_key = _file.read()
        self._encrypted_key = encrypted_key
        self._private_key = private_key

    @property
    def key_file(self):
        if self._key_file:
            return os.path.expandvars(os.path.expanduser(self._key_file))
        return None

    @property
    def key(self):
        if self._private_key:
            return self._private_key

        return self._encrypted_key



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from ocean_keeper.conditions.condition_base import ConditionBase


class EscrowRewardCondition(ConditionBase):
    """Class representing the EscrowReward contract."""
    CONTRACT_NAME = 'EscrowReward'

    def fulfill(self,
                agreement_id,
                amount,
                receiver_address,
                sender_address,
                lock_condition_id,
                release_condition_id,
                account):
        """
        Fulfill the escrow reward condition.

        :param agreement_id: id of the agreement, hex str
        :param amount: Amount of tokens, int
        :param receiver_address: ethereum address of the receiver, hex str
        :param sender_address: ethereum address of the sender, hex str
        :param lock_condition_id: id of the condition, str
        :param release_condition_id: id of the condition, str
        :param account: Account instance
        :return:
        """
        return self._fulfill(
            agreement_id,
            amount,
            receiver_address,
            sender_address,
            lock_condition_id,
            release_condition_id,
            transact={'from': account.address,
                      'passphrase': account.password,
                      'account_key': account.key}
        )

    def hash_values(self, amount, receiver_address, sender_address, lock_condition_id,
                    release_condition_id):
        """
        Hash the values of the escrow reward condition.

        :param amount: Amount of tokens, int
        :param receiver_address: ethereum address of the receiver, hex str
        :param sender_address: ethereum address of the sender, hex str
        :param lock_condition_id: id of the condition, str
        :param release_condition_id: id of the condition, str
        :return: hex str
        """
        return self._hash_values(
            amount,
            receiver_address,
            sender_address,
            lock_condition_id,
            release_condition_id
        )



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from collections import namedtuple

from eth_utils import add_0x_prefix

from ocean_keeper import ContractBase
from ocean_keeper.templates import EscrowAccessSecretStoreTemplate
from ocean_keeper.templates.compute_execution_template import EscrowComputeExecutionTemplate
from ocean_keeper.utils import generate_multi_value_hash

AgreementTemplate = namedtuple(
    'AgreementTemplate',
    ('state', 'owner', 'updated_by', 'block_number_updated', 'condition_types', 'actor_type_ids')
)


class TemplateStoreManager(ContractBase):
    """Class representing the TemplateStoreManager contract."""
    CONTRACT_NAME = 'TemplateStoreManager'
    SERVICE_TO_TEMPLATE_NAME = {
        'access': EscrowAccessSecretStoreTemplate.CONTRACT_NAME,
        'compute': EscrowComputeExecutionTemplate.CONTRACT_NAME
    }

    @staticmethod
    def create_template_id(template_name):
        return generate_multi_value_hash(['string'], [template_name]).hex()

    def get_template_by_name(self, template_name):
        return self.get_template(TemplateStoreManager.create_template_id(template_name))

    def get_template_by_service_type(self, template_type):
        name = self.SERVICE_TO_TEMPLATE_NAME.get(template_type, None)
        if not name:
            return None

        return self.get_template_by_name(name)

    def get_template(self, template_id):
        """
        Get the template for a given template id.

        :param template_id: id of the template, str
        :return:
        """
        template = self.contract_concise.getTemplate(template_id)
        if template and len(template) == 6:
            actor_type_ids = [add_0x_prefix(actor_type_id.hex())
                              for actor_type_id in template[-1]]
            template[-1] = actor_type_ids
            return AgreementTemplate(*template)

        return None

    def get_template_actor_type_ids(self):
        return [add_0x_prefix(actor_type_id.hex())
                for actor_type_id in self.contract_concise.getTemplateActorTypeIds()]

    def get_template_actor_type_id(self, actor_type):
        return add_0x_prefix(self.contract_concise.getTemplateActorTypeId(actor_type).hex())

    def get_template_actor_type_value(self, actor_type_id):
        return self.contract_concise.getTemplateActorTypeValue(actor_type_id)

    def get_template_actor_type_state(self, actor_type_id):
        # 0, 1, or 2 (uninitialized, registered, deregistered
        return self.contract_concise.getTemplateActorTypeState(actor_type_id)

    def propose_template(self, template_id, from_account):
        """Propose a template.

        :param template_id: id of the template, str
        :param from_account: Account
        :return: bool
        """
        tx_hash = self.send_transaction(
            'proposeTemplate',
            (template_id,),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'account_key': from_account.key}
        )
        return self.is_tx_successful(tx_hash)

    def approve_template(self, template_id, from_account):
        """
        Approve a template.

        :param template_id: id of the template, str
        :param from_account: Account
        :return:
        """
        tx_hash = self.send_transaction(
            'approveTemplate',
            (template_id,),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'account_key': from_account.key}
        )
        return self.is_tx_successful(tx_hash)

    def revoke_template(self, template_id, from_account):
        """
        Revoke a template.

        :param template_id: id of the template, str
        :param from_account: Account
        :return: bool
        """
        tx_hash = self.send_transaction(
            'revokeTemplate',
            (template_id,),
            transact={'from': from_account.address,
                      'passphrase': from_account.password,
                      'account_key': from_account.key}
        )
        return self.is_tx_successful(tx_hash)

    def is_template_approved(self, template_id):
        """
        True if the template is approved.

        :param template_id: id of the template, str
        :return: bool
        """
        return self.contract_concise.isTemplateApproved(template_id)

    def get_num_templates(self):
        """
        Return the number of templates on-chain.

        :return: number of templates, int
        """
        return self.contract_concise.getTemplateListSize()



"""Keeper module to call keeper-contracts."""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0


from ocean_keeper.contract_base import ContractBase


class GenericContract(ContractBase):
    """Class for instantiating any contract.

    Contract name is set at time of loading the contract.

    """
